package ru.dbotthepony.kstarbound.defs.world

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import com.google.gson.Gson
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kommons.collect.filterNotNull
import ru.dbotthepony.kommons.gson.consumeNull
import ru.dbotthepony.kommons.gson.stream
import ru.dbotthepony.kommons.gson.value
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.collect.WeightedList
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.builder.JsonFlat
import ru.dbotthepony.kstarbound.json.listAdapter
import ru.dbotthepony.kstarbound.util.random.AbstractPerlinNoise
import java.util.stream.Stream

@JsonFactory
data class BiomePlaceables(
	val grassMod: Registry.Entry<MaterialModifier>? = null,
	val ceilingGrassMod: Registry.Entry<MaterialModifier>? = null,
	val grassModDensity: Double = 0.0,
	val ceilingGrassModDensity: Double = 0.0,
	val items: ImmutableList<DistributionItem> = ImmutableList.of(),
) {
	fun firstTreeVariant(): TreeVariant? {
		return items.stream()
			.flatMap { it.data.itemStream() }
			.map { it as? Tree }
			.filterNotNull()
			.flatMap { it.trees.stream() }
			.findAny()
			.orElse(null)
	}

	// ----------- ITEMS
	@JsonFactory
	data class DistributionItem(
		val priority: Double = 0.0,
		val variants: Int = 1,
		val mode: BiomePlaceablesDefinition.Placement = BiomePlaceablesDefinition.Placement.FLOOR,
		@JsonFlat
		val data: DistributionData,
	)

	abstract class Item {
		abstract val type: BiomePlacementItemType

		abstract fun toJson(): JsonElement

		companion object : TypeAdapterFactory {
			override fun <T : Any> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
				if (Item::class.java.isAssignableFrom(type.rawType)) {
					return object : TypeAdapter<Item>() {
						private val arrays = gson.getAdapter(JsonArray::class.java)
						private val grassVariant = gson.getAdapter(GrassVariant::class.java)
						private val bushVariant = gson.getAdapter(BushVariant::class.java)
						private val trees = gson.listAdapter<TreeVariant>()
						private val objects = gson.getAdapter(PoolTypeToken)

						override fun write(out: JsonWriter, value: Item?) {
							if (value == null)
								out.nullValue()
							else {
								out.beginArray()

								when (value.type) {
									BiomePlacementItemType.MICRO_DUNGEON -> out.value("microDungeon")
									BiomePlacementItemType.TREASURE_BOX_SET -> out.value("treasureBoxSet")
									BiomePlacementItemType.GRASS -> out.value("grass")
									BiomePlacementItemType.BUSH -> out.value("bush")
									BiomePlacementItemType.TREE -> out.value("treePair")
									BiomePlacementItemType.OBJECT -> out.value("objectPool")
								}

								out.value(value.toJson())
								out.endArray()
							}
						}

						override fun read(`in`: JsonReader): Item? {
							if (`in`.consumeNull())
								return null

							`in`.beginArray()

							// sucks that this has to be done manually
							// Also I salute to whoever decided to give different names
							// and different comparison rules for data in *.biome files
							// and world storage data at Chucklefish.
							// Truly our hero here.
							val obj = when (val type = `in`.nextString()) {
								"treasureBoxSet" -> TreasureBox(`in`.nextString())
								"microDungeon" -> MicroDungeon(arrays.read(`in`).stream().map { it.asString }.collect(
									ImmutableSet.toImmutableSet()
								))
								"grass" -> Grass(grassVariant.read(`in`))
								"bush" -> Bush(bushVariant.read(`in`))
								"tree" -> Tree(trees.read(`in`))
								"objectPool" -> Object(objects.read(`in`))
								else -> throw JsonSyntaxException("Unknown biome placement item $type")
							}

							`in`.endArray()
							return obj
						}
					} as TypeAdapter<T>
				}

				return null
			}
		}
	}

	data class MicroDungeon(val microdungeons: ImmutableSet<String> = ImmutableSet.of()) : Item() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.MICRO_DUNGEON

		override fun toJson(): JsonElement {
			return JsonArray().also { j ->
				microdungeons.forEach { j.add(JsonPrimitive(it)) }
			}
		}
	}

	data class TreasureBox(val pool: String) : Item() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.TREASURE_BOX_SET

		override fun toJson(): JsonElement {
			return JsonPrimitive(pool)
		}
	}

	data class Grass(val value: GrassVariant) : Item() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.GRASS

		override fun toJson(): JsonElement {
			return Starbound.gson.toJsonTree(value)
		}
	}

	data class Bush(val value: BushVariant) : Item() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.BUSH

		override fun toJson(): JsonElement {
			return Starbound.gson.toJsonTree(value)
		}
	}

	data class Tree(val trees: ImmutableList<TreeVariant>) : Item() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.TREE

		override fun toJson(): JsonElement {
			return Starbound.gson.toJsonTree(trees, TypeToken.getParameterized(
				ImmutableList::class.java,
				TreeVariant::class.java
			).type)
		}
	}

	private object PoolTypeToken : TypeToken<WeightedList<Pair<String, JsonElement>>>()

	// This structure sucks, but at least it allows unique parameters per
	// each object (lmao, whos gonna write world json by hand anyway????
	// considering this is world generation data.)
	data class Object(val pool: WeightedList<Pair<String, JsonElement>>) : Item() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.OBJECT

		override fun toJson(): JsonElement {
			return Starbound.gson.toJsonTree(pool, PoolTypeToken.type)
		}
	}

	// -------- DISTRIBUTION
	abstract class DistributionData {
		abstract val type: BiomePlacementDistributionType

		abstract fun itemStream(): Stream<Item>
	}

	@JsonFactory
	data class RandomDistribution(
		val blockSeed: Long,
		val randomItems: ImmutableList<Item>,
	) : DistributionData() {
		override val type: BiomePlacementDistributionType
			get() = BiomePlacementDistributionType.RANDOM

		override fun itemStream(): Stream<Item> {
			return randomItems.stream()
		}
	}

	@JsonFactory
	data class PeriodicDistribution(
		val modulus: Int,
		val modulusOffset: Int,
		val densityFunction: AbstractPerlinNoise,
		val modulusDistortion: AbstractPerlinNoise,
		val weightedItems: ImmutableList<Pair<Item, AbstractPerlinNoise>>,
	) : DistributionData() {
		override val type: BiomePlacementDistributionType
			get() = BiomePlacementDistributionType.PERIODIC

		override fun itemStream(): Stream<Item> {
			return weightedItems.stream().map { it.first }
		}
	}
}