package ru.dbotthepony.kstarbound.lua

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import it.unimi.dsi.fastutil.longs.Long2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectIterators
import org.classdump.luna.ByteString
import org.classdump.luna.Conversions
import org.classdump.luna.LuaRuntimeException
import org.classdump.luna.Table
import org.classdump.luna.TableFactory
import org.classdump.luna.impl.NonsuspendableFunctionException
import org.classdump.luna.lib.ArgumentIterator
import org.classdump.luna.lib.BadArgumentException
import org.classdump.luna.runtime.AbstractFunction0
import org.classdump.luna.runtime.AbstractFunction1
import org.classdump.luna.runtime.AbstractFunction2
import org.classdump.luna.runtime.AbstractFunction3
import org.classdump.luna.runtime.AbstractFunction4
import org.classdump.luna.runtime.AbstractFunctionAnyArg
import org.classdump.luna.runtime.ExecutionContext
import org.classdump.luna.runtime.LuaFunction
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kommons.util.IStruct3i
import ru.dbotthepony.kommons.util.IStruct4i
import ru.dbotthepony.kstarbound.json.InternedJsonElementAdapter

fun ArgumentIterator.nextOptionalFloat(): Double? {
	if (hasNext()) {
		return nextFloat()
	} else {
		return null
	}
}

fun ArgumentIterator.nextOptionalInteger(): Long? {
	if (hasNext()) {
		return nextInteger()
	} else {
		return null
	}
}

operator fun Table.set(index: Any, value: Any?) {
	rawset(index, value)
}

@Deprecated("Trying to set JSON to Lua table", level = DeprecationLevel.ERROR)
operator fun Table.set(index: Any, value: JsonElement?) {
	rawset(index, value)
}

operator fun Table.set(index: Long, value: Any?) {
	rawset(index, value)
}

@Deprecated("Trying to set JSON to Lua table", level = DeprecationLevel.ERROR)
operator fun Table.set(index: Long, value: JsonElement?) {
	rawset(index, value)
}

operator fun Table.set(index: Int, value: Any?) {
	rawset(index.toLong(), value)
}

@Deprecated("Trying to set JSON to Lua table", level = DeprecationLevel.ERROR)
operator fun Table.set(index: Int, value: JsonElement?) {
	rawset(index.toLong(), value)
}

operator fun Table.get(index: Any): Any? = rawget(index)
operator fun Table.get(index: Long): Any? = rawget(index)
operator fun Table.get(index: Int): Any? = rawget(index.toLong())

operator fun Table.iterator(): Iterator<Map.Entry<Any, Any>> {
	var key: Any? = initialKey() ?: return ObjectIterators.emptyIterator()
	data class Pair(override val key: Any, override val value: Any) : Map.Entry<Any, Any>

	return object : Iterator<Map.Entry<Any, Any>> {
		override fun hasNext(): Boolean {
			return key != null
		}

		override fun next(): Map.Entry<Any, Any> {
			val ikey = key ?: throw NoSuchElementException()
			val value = get(ikey)!!
			key = successorKeyOf(ikey)
			return Pair(ikey, value)
		}
	}
}

fun Table.toJson(): JsonElement {
	val arrayValues = Long2ObjectAVLTreeMap<Any>()
	val hashValues = HashMap<Any, Any>()

	for ((k, v) in this) {
		if (k is Number) {
			arrayValues[k.toLong()] = v
		} else {
			hashValues[k] = v
		}
	}

	if (hashValues.isEmpty()) {
		return JsonArray(arrayValues.size).also {
			for (v in arrayValues.values) {
				if (v is ByteString) {
					it.add(JsonPrimitive(v.decode()))
				} else if (v is Number) {
					it.add(JsonPrimitive(v))
				} else if (v is Boolean) {
					it.add(InternedJsonElementAdapter.of(v))
				} else if (v is Table) {
					it.add(v.toJson())
				}
			}
		}
	} else {
		return JsonObject().also {
			for ((k, v) in arrayValues) {
				if (v is ByteString) {
					it.add(k.toString(), JsonPrimitive(v.decode()))
				} else if (v is Number) {
					it.add(k.toString(), JsonPrimitive(v))
				} else if (v is Boolean) {
					it.add(k.toString(), InternedJsonElementAdapter.of(v))
				} else if (v is Table) {
					it.add(k.toString(), v.toJson())
				}
			}

			for ((k, v) in hashValues) {
				if (v is ByteString) {
					it.add(k.toString(), JsonPrimitive(v.decode()))
				} else if (v is Number) {
					it.add(k.toString(), JsonPrimitive(v))
				} else if (v is Boolean) {
					it.add(k.toString(), InternedJsonElementAdapter.of(v))
				} else if (v is Table) {
					it.add(k.toString(), v.toJson())
				}
			}
		}
	}
}

fun Table.toJsonObject(): JsonObject {
	return JsonObject().also {
		for ((k, v) in this) {
			if (v is ByteString) {
				it.add(k.toString(), JsonPrimitive(v.decode()))
			} else if (v is Number) {
				it.add(k.toString(), JsonPrimitive(v))
			} else if (v is Boolean) {
				it.add(k.toString(), InternedJsonElementAdapter.of(v))
			} else if (v is Table) {
				it.add(k.toString(), v.toJson())
			}
		}
	}
}

fun TableFactory.from(value: JsonElement?): Any? {
	when (value) {
		is JsonPrimitive -> {
			if (value.isNumber) {
				return value.asDouble
			} else if (value.isString) {
				return ByteString.of(value.asString)
			} else if (value.isBoolean) {
				return value.asBoolean
			} else {
				throw RuntimeException("unreachable code")
			}
		}

		is JsonNull -> return null
		is JsonArray -> return from(value)
		is JsonObject -> return from(value)
		null -> return null
		else -> throw RuntimeException(value::class.qualifiedName)
	}
}

fun TableFactory.from(value: JsonObject): Table {
	val table = newTable(0, value.size())

	for ((k, v) in value.entrySet()) {
		table.rawset(Conversions.normaliseKey(k), from(v))
	}

	return table
}

fun TableFactory.from(value: JsonArray): Table {
	val table = newTable(value.size(), 0)

	for ((k, v) in value.withIndex()) {
		table.rawset(Conversions.normaliseKey(k + 1), from(v))
	}

	return table
}

fun TableFactory.fromCollection(value: Collection<JsonElement?>): Table {
	val table = newTable(value.size, 0)

	for ((k, v) in value.withIndex()) {
		table.rawset(Conversions.normaliseKey(k + 1), from(v))
	}

	return table
}

fun TableFactory.from(value: IStruct2i): Table {
	return newTable(2, 0).also {
		it.rawset(1L, value.component1())
		it.rawset(2L, value.component2())
	}
}

fun TableFactory.from(value: IStruct3i): Table {
	return newTable(3, 0).also {
		it.rawset(1L, value.component1())
		it.rawset(2L, value.component2())
		it.rawset(3L, value.component3())
	}
}

fun TableFactory.from(value: IStruct4i): Table {
	return newTable(3, 0).also {
		it.rawset(1L, value.component1())
		it.rawset(2L, value.component2())
		it.rawset(3L, value.component3())
		it.rawset(4L, value.component4())
	}
}

fun TableFactory.from(value: Collection<Any>): Table {
	return newTable(value.size, 0).also {
		for ((i, v) in value.withIndex()) {
			it.rawset(i + 1L, v)
		}
	}
}

@Deprecated("Function is a stub")
fun luaStub(message: String = "not yet implemented"): LuaFunction<Any?, Any?, Any?, Any?, Any?> {
	return object : LuaFunction<Any?, Any?, Any?, Any?, Any?>() {
		override fun resume(context: ExecutionContext?, suspendedState: Any?) {
			throw NonsuspendableFunctionException(this::class.java)
		}

		override fun invoke(context: ExecutionContext?) {
			throw LuaRuntimeException("NYI: $message")
		}

		override fun invoke(context: ExecutionContext?, arg1: Any?) {
			throw LuaRuntimeException("NYI: $message")
		}

		override fun invoke(context: ExecutionContext?, arg1: Any?, arg2: Any?) {
			throw LuaRuntimeException("NYI: $message")
		}

		override fun invoke(context: ExecutionContext?, arg1: Any?, arg2: Any?, arg3: Any?) {
			throw LuaRuntimeException("NYI: $message")
		}

		override fun invoke(context: ExecutionContext?, arg1: Any?, arg2: Any?, arg3: Any?, arg4: Any?) {
			throw LuaRuntimeException("NYI: $message")
		}

		override fun invoke(context: ExecutionContext?, arg1: Any?, arg2: Any?, arg3: Any?, arg4: Any?, arg5: Any?) {
			throw LuaRuntimeException("NYI: $message")
		}

		override fun invoke(context: ExecutionContext?, args: Array<out Any>?) {
			throw LuaRuntimeException("NYI: $message")
		}
	}
}

fun luaFunction0String(name: String, callable: (String) -> Any?): LuaFunction<Any?, *, *, *, *> {
	return object : AbstractFunction1<Any?>() {
		override fun resume(context: ExecutionContext, suspendedState: Any) {
			throw NonsuspendableFunctionException(this::class.java)
		}

		override fun invoke(context: ExecutionContext, arg: Any?) {
			if (arg !is ByteString)
				throw BadArgumentException(1, name, "string expected, got ${if (arg == null) "nil" else arg::class.qualifiedName}")

			val result = callable.invoke(arg.decode())

			if (result != null && result !== Unit) {
				context.returnBuffer.setTo(result)
			}
		}
	}
}

fun luaFunctionN(name: String, callable: (ExecutionContext, ArgumentIterator) -> Any?): LuaFunction<Any, Any, Any, Any, Any> {
	return object : AbstractFunctionAnyArg() {
		override fun resume(context: ExecutionContext, suspendedState: Any) {
			throw NonsuspendableFunctionException(this::class.java)
		}

		override fun invoke(context: ExecutionContext, args: Array<out Any>) {
			val result = callable.invoke(context, ArgumentIterator.of(context, name, args))

			if (result != null && result !== Unit) {
				context.returnBuffer.setTo(result)
			}
		}
	}
}

fun luaFunctionNS(name: String, callable: (ExecutionContext, ArgumentIterator) -> StateMachine): LuaFunction<Any, Any, Any, Any, Any> {
	return object : AbstractFunctionAnyArg() {
		override fun resume(context: ExecutionContext, suspendedState: Any) {
			(suspendedState as StateMachine).run(context)
		}

		override fun invoke(context: ExecutionContext, args: Array<out Any>) {
			callable.invoke(context, ArgumentIterator.of(context, name, args)).run(context)
		}
	}
}

fun luaFunction(callable: (ExecutionContext) -> Unit): LuaFunction<*, *, *, *, *> {
	return object : AbstractFunction0() {
		override fun resume(context: ExecutionContext, suspendedState: Any) {
			throw NonsuspendableFunctionException(this::class.java)
		}

		override fun invoke(context: ExecutionContext) {
			callable.invoke(context)
		}
	}
}

fun <T> luaFunction(callable: (ExecutionContext, T) -> Unit): LuaFunction<T, *, *, *, *> {
	return object : AbstractFunction1<T>() {
		override fun resume(context: ExecutionContext, suspendedState: Any) {
			throw NonsuspendableFunctionException(this::class.java)
		}

		override fun invoke(context: ExecutionContext, arg1: T) {
			callable.invoke(context, arg1)
		}
	}
}

fun <T, T2> luaFunction(callable: (ExecutionContext, T, T2) -> Unit): LuaFunction<T, T2, *, *, *> {
	return object : AbstractFunction2<T, T2>() {
		override fun resume(context: ExecutionContext, suspendedState: Any) {
			throw NonsuspendableFunctionException(this::class.java)
		}

		override fun invoke(context: ExecutionContext, arg1: T, arg2: T2) {
			callable.invoke(context, arg1, arg2)
		}
	}
}

fun <T, T2, T3> luaFunction(callable: (ExecutionContext, T, T2, T3) -> Unit): LuaFunction<T, T2, T3, *, *> {
	return object : AbstractFunction3<T, T2, T3>() {
		override fun resume(context: ExecutionContext, suspendedState: Any) {
			throw NonsuspendableFunctionException(this::class.java)
		}

		override fun invoke(context: ExecutionContext, arg1: T, arg2: T2, arg3: T3) {
			callable.invoke(context, arg1, arg2, arg3)
		}
	}
}

fun <T, T2, T3, T4> luaFunction(callable: (ExecutionContext, T, T2, T3, T4) -> Unit): LuaFunction<T, T2, T3, T4, *> {
	return object : AbstractFunction4<T, T2, T3, T4>() {
		override fun resume(context: ExecutionContext, suspendedState: Any) {
			throw NonsuspendableFunctionException(this::class.java)
		}

		override fun invoke(context: ExecutionContext, arg1: T, arg2: T2, arg3: T3, arg4: T4) {
			callable.invoke(context, arg1, arg2, arg3, arg4)
		}
	}
}
