package ru.dbotthepony.kstarbound.server

import com.google.gson.JsonObject
import io.netty.channel.ChannelHandlerContext
import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.ints.Int2LongOpenHashMap
import it.unimi.dsi.fastutil.ints.Int2ObjectMaps
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import it.unimi.dsi.fastutil.objects.ObjectLinkedOpenHashSet
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.io.ByteKey
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.client.network.packets.ChunkCellsPacket
import ru.dbotthepony.kstarbound.defs.WarpAlias
import ru.dbotthepony.kstarbound.network.Connection
import ru.dbotthepony.kstarbound.network.ConnectionSide
import ru.dbotthepony.kstarbound.network.ConnectionType
import ru.dbotthepony.kstarbound.network.IServerPacket
import ru.dbotthepony.kstarbound.network.packets.ClientContextUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.EntityCreatePacket
import ru.dbotthepony.kstarbound.network.packets.EntityUpdateSetPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.LegacyTileArrayUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.LegacyTileUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.PlayerWarpResultPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.ServerDisconnectPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.TileDamageUpdatePacket
import ru.dbotthepony.kstarbound.server.world.WorldStorage
import ru.dbotthepony.kstarbound.server.world.LegacyWorldStorage
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.IChunkListener
import ru.dbotthepony.kstarbound.world.TileHealth
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.player.PlayerEntity
import java.io.DataOutputStream
import java.util.HashMap
import java.util.concurrent.ConcurrentLinkedQueue
import kotlin.properties.Delegates

// serverside part of connection
class ServerConnection(val server: StarboundServer, type: ConnectionType) : Connection(ConnectionSide.SERVER, type) {
	var world: ServerWorld? = null
	var worldStartAcknowledged = false

	private val tasks = ConcurrentLinkedQueue<ServerWorld.() -> Unit>()

	// packets which interact with world must be
	// executed on world's thread
	fun enqueue(task: ServerWorld.() -> Unit) {
		tasks.add(task)
	}

	lateinit var shipWorld: ServerWorld
		private set

	var skyVersion = 0L

	init {
		connectionID = server.channels.nextConnectionID()

		rpc.add("team.fetchTeamStatus") {
			JsonObject()
		}
	}

	override fun toString(): String {
		val channel = if (hasChannel) channel.remoteAddress().toString() else "<no channel>"
		val ship = if (::shipWorld.isInitialized) shipWorld.toString() else "<no shipworld>"
		return "ServerConnection[ID=$connectionID channel=$channel / $ship]"
	}

	private val shipChunks = HashMap<ByteKey, KOptional<ByteArray>>()
	private val modifiedShipChunks = ObjectOpenHashSet<ByteKey>()
	var shipChunkSource by Delegates.notNull<WorldStorage>()
		private set

	override fun setupLegacy() {
		super.setupLegacy()
		shipChunkSource = LegacyWorldStorage.memory(shipChunks)
	}

	override fun setupNative() {
		super.setupNative()
		shipChunkSource = WorldStorage.EMPTY
	}

	fun receiveShipChunks(chunks: Map<ByteKey, KOptional<ByteArray>>) {
		check(shipChunks.isEmpty()) { "Already has ship chunks" }
		shipChunks.putAll(chunks)
	}

	private val tickets = HashMap<ChunkPos, Ticket>()
	private val pendingSend = ObjectLinkedOpenHashSet<ChunkPos>()

	private inner class Ticket(val ticket: ServerWorld.ITicket, val pos: ChunkPos) : IChunkListener {
		override fun onEntityAdded(entity: AbstractEntity) {}
		override fun onEntityRemoved(entity: AbstractEntity) {}

		override fun onCellChanges(x: Int, y: Int, cell: ImmutableCell) {
			if (pos !in pendingSend) {
				send(LegacyTileUpdatePacket(pos.tile + Vector2i(x, y), cell.toLegacyNet()))
			}
		}

		override fun onTileHealthUpdate(x: Int, y: Int, isBackground: Boolean, health: TileHealth) {
			// let's hope nothing bad happens from referencing live data
			send(TileDamageUpdatePacket(pos.tileX + x, pos.tileY + y, isBackground, health))
		}
	}

	override fun flush() {
		if (isConnected) {
			val entries = rpc.write()

			if (entries != null || modifiedShipChunks.isNotEmpty()) {
				channel.write(ClientContextUpdatePacket(
					entries ?: listOf(),
					KOptional(modifiedShipChunks.associateWith { shipChunks[it]!! }),
					KOptional(ByteArrayList())))

				modifiedShipChunks.clear()
			}
		}

		super.flush()
	}

	fun onLeaveWorld() {
		tasks.clear()
		tickets.values.forEach { it.ticket.cancel() }
		tickets.clear()
		pendingSend.clear()
		playerEntity = null
	}

	override fun onChannelClosed() {
		playerEntity = null

		super.onChannelClosed()
		server.channels.freeConnectionID(connectionID)
		server.channels.connections.remove(this)
		server.freeNickname(nickname)

		announceDisconnect("Connection to remote host is lost.")

		if (::shipWorld.isInitialized) {
			shipWorld.close()
		}

		if (countedTowardsPlayerCount) {
			countedTowardsPlayerCount = false
			server.channels.decrementPlayerCount()
		}
	}

	private var announcedDisconnect = false

	private fun announceDisconnect(reason: String) {
		if (!announcedDisconnect && nickname.isNotBlank()) {
			if (reason.isBlank()) {
				server.chat.systemMessage("Player '$nickname' disconnected")
			} else {
				server.chat.systemMessage("Player '$nickname' disconnected ($reason)")
			}

			announcedDisconnect = true
		}
	}

	override fun disconnect(reason: String) {
		announceDisconnect(reason)

		if (channel.isOpen) {
			// send pending updates
			flush()
		}

		world?.removePlayer(this)
		world = null
		tickets.values.forEach { it.ticket.cancel() }
		tickets.clear()
		tasks.clear()
		pendingSend.clear()

		if (::shipWorld.isInitialized) {
			shipWorld.close()
		}

		if (channel.isOpen) {
			// say goodbye
			channel.write(ServerDisconnectPacket(reason))
			channel.flush()
			channel.close()
		}
	}

	private val entityVersions = Int2LongOpenHashMap()

	init {
		entityVersions.defaultReturnValue(-1L)
	}

	fun isTracking(pos: ChunkPos): Boolean {
		return pos in tickets
	}

	fun tickWorld() {
		val world = world!!

		run {
			var next = tasks.poll()

			while (next != null) {
				next.invoke(world)
				next = tasks.poll()
			}
		}

		playerEntity = world.entities[playerID.get()] as? PlayerEntity

		run {
			val newTrackedChunks = ObjectArraySet<ChunkPos>()

			for (region in trackingTileRegions()) {
				newTrackedChunks.addAll(world.geometry.tileRegion2Chunks(region))
			}

			val itr = tickets.entries.iterator()

			for ((pos, ticket) in itr) {
				if (pos !in newTrackedChunks) {
					pendingSend.remove(pos)
					ticket.ticket.cancel()
					itr.remove()
				}
			}

			for (pos in newTrackedChunks) {
				if (pos !in tickets) {
					val ticket = world.permanentChunkTicket(pos)
					val thisTicket = Ticket(ticket, pos)
					tickets[pos] = thisTicket
					ticket.listener = thisTicket
					pendingSend.add(pos)
				}
			}
		}

		run {
			val itr = pendingSend.iterator()

			for (pos in itr) {
				val chunk = world.chunkMap[pos] ?: continue

				if (isLegacy) {
					send(LegacyTileArrayUpdatePacket(chunk))
					chunk.tileDamagePackets().forEach { send(it) }
				} else {
					send(ChunkCellsPacket(chunk))
				}

				itr.remove()
			}
		}

		for ((id, entity) in world.entities) {
			if (entity.connectionID != connectionID && entity is PlayerEntity) {
				if (entityVersions.get(id) == -1L) {
					// never networked
					val initial = FastByteArrayOutputStream()
					entity.writeNetwork(DataOutputStream(initial), isLegacy)
					val (data, version) = entity.networkGroup.write(isLegacy = isLegacy)

					entityVersions.put(id, version)

					send(EntityCreatePacket(
						entity.type,
						ByteArrayList.wrap(initial.array, initial.length),
						data,
						entity.entityID
					))
				} else {
					val (data, version) = entity.networkGroup.write(remoteVersion = entityVersions.get(id), isLegacy = isLegacy)
					entityVersions.put(id, version)
					send(EntityUpdateSetPacket(entity.connectionID, Int2ObjectMaps.singleton(entity.entityID, data)))
				}
			}
		}
	}

	override fun channelRead(ctx: ChannelHandlerContext, msg: Any) {
		if (msg is IServerPacket) {
			try {
				msg.play(this)
			} catch (err: Throwable) {
				LOGGER.error("Failed to handle serverbound packet $msg", err)
				disconnect("Incoming packet caused an exception: $err")
			}
		} else {
			LOGGER.error("Unknown serverbound packet type $msg")
			disconnect("Unknown serverbound packet type $msg")
		}
	}

	private var countedTowardsPlayerCount = false

	override fun inGame() {
		server.chat.systemMessage("Player '$nickname' connected")
		countedTowardsPlayerCount = true
		server.channels.incrementPlayerCount()

		if (!isLegacy) {
			server.playerInGame(this)
		} else {
			LOGGER.info("Initializing ship world for $this")

			ServerWorld.load(server, shipChunkSource).thenAccept {
				shipWorld = it
				shipWorld.thread.start()
				send(PlayerWarpResultPacket(true, WarpAlias.OwnShip, false))
				shipWorld.acceptPlayer(this).thenAccept {
					for (conn in server.channels.connections) {
						if (conn.isLegacy && conn !== this) {
							conn.shipWorld.acceptPlayer(this)
							break
						}
					}
				}.exceptionally {
					LOGGER.error("Shipworld of $this rejected to accept its owner", it)
					disconnect("Shipworld rejected player warp request: $it")
					null
				}
			}.exceptionally {
				LOGGER.error("Error while initializing shipworld for $this", it)
				disconnect("Error while initializing shipworld for player: $it")
				null
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
