package ru.dbotthepony.kstarbound.server.world

import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.ints.IntArraySet
import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.network.packets.JoinWorldPacket
import ru.dbotthepony.kstarbound.defs.tile.TileDamage
import ru.dbotthepony.kstarbound.defs.tile.TileDamageResult
import ru.dbotthepony.kstarbound.defs.world.WorldStructure
import ru.dbotthepony.kstarbound.defs.world.WorldTemplate
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.network.IPacket
import ru.dbotthepony.kstarbound.network.packets.StepUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.CentralStructureUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.WorldStartPacket
import ru.dbotthepony.kstarbound.server.StarboundServer
import ru.dbotthepony.kstarbound.server.ServerConnection
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.ExecutionSpinner
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.IChunkListener
import ru.dbotthepony.kstarbound.world.TileHealth
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.WorldGeometry
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import java.util.Collections
import java.util.concurrent.CompletableFuture
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.RejectedExecutionException
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.atomic.AtomicInteger
import java.util.concurrent.locks.LockSupport
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Consumer
import java.util.function.Supplier
import kotlin.concurrent.withLock

class ServerWorld private constructor(
	val server: StarboundServer,
	template: WorldTemplate,
	val storage: WorldStorage,
) : World<ServerWorld, ServerChunk>(template) {
	init {
		if (server.isClosed)
			throw RuntimeException()

		server.worlds.add(this)
	}

	private val internalPlayers = CopyOnWriteArrayList<ServerConnection>()
	val players: List<ServerConnection> = Collections.unmodifiableList(internalPlayers)

	private fun doAcceptPlayer(player: ServerConnection) {
		if (player in internalPlayers)
			throw IllegalStateException("$player is already in $this")

		internalPlayers.add(player)
		player.onLeaveWorld()
		player.world?.removePlayer(player)
		player.world = this
		player.worldStartAcknowledged = false

		if (player.isLegacy) {
			val (skyData, skyVersion) = sky.networkedGroup.write(isLegacy = true)
			player.skyVersion = skyVersion

			player.send(WorldStartPacket(
				templateData = template.toJson(true),
				skyData = skyData.toByteArray(),
				weatherData = ByteArray(0),
				playerStart = playerSpawnPosition,
				playerRespawn = playerSpawnPosition,
				respawnInWorld = respawnInWorld,
				dungeonGravity = mapOf(),
				dungeonBreathable = mapOf(),
				protectedDungeonIDs = protectedDungeonIDs,
				worldProperties = properties.deepCopy(),
				connectionID = player.connectionID,
				localInterpolationMode = false,
			))

			player.sendAndFlush(CentralStructureUpdatePacket(Starbound.gson.toJsonTree(centralStructure)))
		} else {
			player.sendAndFlush(JoinWorldPacket(this))
		}
	}

	fun acceptPlayer(player: ServerConnection): CompletableFuture<Unit> {
		check(!isClosed.get()) { "$this is invalid" }
		unpause()

		try {
			return CompletableFuture.supplyAsync(Supplier { doAcceptPlayer(player) }, mailbox)
		} catch (err: RejectedExecutionException) {
			return CompletableFuture.failedFuture(err)
		}
	}

	private fun doRemovePlayer(player: ServerConnection): Boolean {
		if (internalPlayers.remove(player)) {
			val itr = entities.int2ObjectEntrySet().iterator()

			for ((id, entity) in itr) {
				if (id in player.entityIDRange) {
					entity.remove()
				}
			}

			return true
		}

		return false
	}

	fun removePlayer(player: ServerConnection): CompletableFuture<Boolean> {
		check(!isClosed.get()) { "$this is invalid" }

		try {
			return CompletableFuture.supplyAsync(Supplier { doRemovePlayer(player) }, mailbox)
		} catch (err: RejectedExecutionException) {
			return CompletableFuture.completedFuture(false)
		}
	}

	val spinner = ExecutionSpinner(mailbox::executeQueuedTasks, ::spin, Starbound.TIMESTEP_NANOS)
	val thread = Thread(spinner, "Starbound Server World Thread")
	val ticketListLock = ReentrantLock()

	private val isClosed = AtomicBoolean()

	init {
		thread.isDaemon = true
	}

	fun pause() {
		if (!isClosed.get()) spinner.pause()
	}

	fun unpause() {
		if (!isClosed.get()) spinner.unpause()
	}

	override fun close() {
		if (isClosed.compareAndSet(false, true)) {
			LOGGER.info("Shutting down $this")

			super.close()
			spinner.unpause()

			lock.withLock {
				internalPlayers.forEach {
					it.world = null
				}
			}

			server.worlds.remove(this)
			LockSupport.unpark(thread)
		}
	}

	private fun spin(): Boolean {
		if (isClosed.get()) return false

		try {
			tick()
			return true
		} catch (err: Throwable) {
			LOGGER.fatal("Exception in world tick loop", err)
			close()
			return false
		}
	}

	override val isRemote: Boolean
		get() = false

	override fun isSameThread(): Boolean {
		return Thread.currentThread() === thread
	}

	fun damageTiles(positions: Collection<IStruct2i>, isBackground: Boolean, sourcePosition: Vector2d, damage: TileDamage, source: AbstractEntity? = null): TileDamageResult {
		if (damage.amount <= 0.0)
			return TileDamageResult.NONE

		val actualPositions = positions.stream().map { geometry.wrap(it) }.distinct().toList()
		var topMost = TileDamageResult.NONE

		for (pos in actualPositions) {
			val chunk = chunkMap[geometry.chunkFromCell(pos)] ?: continue
			topMost = topMost.coerceAtLeast(chunk.damageTile(pos - chunk.pos.tile, isBackground, sourcePosition, damage, source))
		}

		return topMost
	}

	override fun tickInner() {
		val packet = StepUpdatePacket(ticks)

		internalPlayers.forEach {
			if (!isClosed.get() && it.worldStartAcknowledged && it.channel.isOpen) {
				it.send(packet)

				try {
					it.tickWorld()
				} catch (err: Throwable) {
					LOGGER.error("Exception while ticking player $it", err)
					//it.disconnect("Exception while ticking player: $err")
				}
			}
		}

		ticketListLock.withLock {
			ticketLists.removeIf {
				val valid = it.tick()

				if (!valid) {
					val removed = ticketMap.remove(it.pos.toLong())
					check(removed == it) { "Expected to remove $it, but removed $removed" }

					val chunk = chunkMap[it.pos]

					if (chunk != null) {
						val unloadable = chunk.entities.filter { it.isApplicableForUnloading }

						storage.saveCells(it.pos, chunk.copyCells())
						storage.saveEntities(it.pos, unloadable)

						unloadable.forEach {
							it.remove()
						}

						chunkMap.remove(it.pos)
					}
				}

				!valid
			}
		}
	}

	override fun broadcast(packet: IPacket) {
		internalPlayers.forEach {
			it.send(packet)
		}
	}

	override fun chunkFactory(pos: ChunkPos): ServerChunk {
		return ServerChunk(this, pos)
	}

	private val ticketMap = Long2ObjectOpenHashMap<TicketList>()
	private val ticketLists = ArrayList<TicketList>()

	private fun getTicketList(pos: ChunkPos): TicketList {
		return ticketMap.computeIfAbsent(geometry.wrapToLong(pos), Long2ObjectFunction { TicketList(it) })
	}

	fun permanentChunkTicket(pos: ChunkPos): ITicket {
		lock.withLock {
			return getTicketList(pos).Ticket()
		}
	}

	fun temporaryChunkTicket(pos: ChunkPos, time: Int): ITicket {
		require(time > 0) { "Invalid ticket time: $time" }

		lock.withLock {
			return getTicketList(pos).TimedTicket(time)
		}
	}

	override fun onChunkCreated(chunk: ServerChunk) {
		ticketMap[chunk.pos.toLong()]?.let { chunk.addListener(it) }
	}

	override fun onChunkRemoved(chunk: ServerChunk) {
		ticketMap[chunk.pos.toLong()]?.let { chunk.removeListener(it) }
	}

	interface ITicket {
		fun cancel()
		val isCanceled: Boolean
		val pos: ChunkPos
		val id: Int
		val chunk: ServerChunk?
		var listener: IChunkListener?
	}

	interface ITimedTicket : ITicket, Comparable<ITimedTicket> {
		val timeRemaining: Int
		fun prolong(ticks: Int)

		override fun compareTo(other: ITimedTicket): Int {
			val cmp = timeRemaining.compareTo(other.timeRemaining)
			if (cmp != 0) return cmp
			return id.compareTo(other.id)
		}
	}

	private inner class TicketList(val pos: ChunkPos) : IChunkListener {
		constructor(pos: Long) : this(ChunkPos(pos))

		private var first = true
		private val permanent = ArrayList<Ticket>()
		private val temporary = ObjectAVLTreeSet<TimedTicket>()
		private var ticks = 0
		private var nextTicketID = AtomicInteger()

		val isValid: Boolean
			get() = temporary.isNotEmpty() || permanent.isNotEmpty()

		fun tick(): Boolean {
			ticks++

			while (temporary.isNotEmpty() && temporary.first().timeRemaining <= 0) {
				val ticket = temporary.first()
				ticket.isCanceled = true
				temporary.remove(ticket)
			}

			return temporary.isNotEmpty() || permanent.isNotEmpty()
		}

		override fun onEntityAdded(entity: AbstractEntity) {
			permanent.forEach { it.listener?.onEntityAdded(entity) }
			temporary.forEach { it.listener?.onEntityAdded(entity) }
		}

		override fun onEntityRemoved(entity: AbstractEntity) {
			permanent.forEach { it.listener?.onEntityRemoved(entity) }
			temporary.forEach { it.listener?.onEntityRemoved(entity) }
		}

		override fun onCellChanges(x: Int, y: Int, cell: ImmutableCell) {
			permanent.forEach { it.listener?.onCellChanges(x, y, cell) }
			temporary.forEach { it.listener?.onCellChanges(x, y, cell) }
		}

		override fun onTileHealthUpdate(x: Int, y: Int, isBackground: Boolean, health: TileHealth) {
			permanent.forEach { it.listener?.onTileHealthUpdate(x, y, isBackground, health) }
			temporary.forEach { it.listener?.onTileHealthUpdate(x, y, isBackground, health) }
		}

		abstract inner class AbstractTicket : ITicket {
			final override val id: Int = nextTicketID.getAndIncrement()
			final override val pos: ChunkPos
				get() = this@TicketList.pos

			final override var isCanceled: Boolean = false
			private var loadFuture: CompletableFuture<*>? = null

			fun init() {
				if (first) {
					first = false

					if (geometry.x.inBoundsChunk(pos.x) && geometry.y.inBoundsChunk(pos.y)) {
						ticketListLock.withLock {
							ticketLists.add(this@TicketList)
						}

						val existing = chunkMap[pos]

						if (existing == null) {
							loadFuture = storage.loadCells(pos).thenAccept { tiles ->
								if (!tiles.isPresent) return@thenAccept

								storage.loadEntities(pos).thenAcceptAsync(Consumer { ents ->
									val chunk = chunkMap.compute(pos) ?: return@Consumer
									chunk.loadCells(tiles.value)

									ents.ifPresent {
										for (obj in it) {
											obj.joinWorld(this@ServerWorld)
										}
									}
								}, mailbox)
							}
						} else {
							existing.addListener(this@TicketList)
						}
					}
				} else {
					chunkMap[pos]?.addListener(this@TicketList)
				}
			}

			final override fun cancel() {
				if (isCanceled) return

				lock.withLock {
					if (isCanceled) return
					isCanceled = true
					chunk?.entities?.forEach { e -> listener?.onEntityRemoved(e) }
					loadFuture?.cancel(false)
					listener = null
					cancel0()
				}
			}

			protected abstract fun cancel0()
			final override val chunk: ServerChunk?
				get() = chunkMap[pos]

			final override var listener: IChunkListener? = null
				set(value) {
					if (field != value) {
						val chunk = chunk
						chunk?.entities?.forEach { e -> field?.onEntityRemoved(e) }
						chunk?.entities?.forEach { e -> value?.onEntityAdded(e) }
						field = value
					}
				}
		}

		inner class Ticket : AbstractTicket() {
			init {
				permanent.add(this)
				init()
			}

			override fun cancel0() {
				permanent.remove(this)
			}
		}

		inner class TimedTicket(expiresAt: Int) : AbstractTicket(), ITimedTicket {
			var expiresAt = expiresAt + ticks

			override val timeRemaining: Int
				get() = (expiresAt - ticks).coerceAtLeast(0)

			init {
				temporary.add(this)
				init()
			}

			override fun cancel0() {
				temporary.remove(this)
			}

			override fun prolong(ticks: Int) {
				if (ticks == 0 || isCanceled) return

				lock.withLock {
					if (isCanceled) return

					temporary.remove(this)
					expiresAt += ticks
					if (timeRemaining > 0) temporary.add(this)
				}
			}
		}
	}

	@JsonFactory
	data class MetadataJson(
		val playerStart: Vector2d,
		val respawnInWorld: Boolean,
		val adjustPlayerStart: Boolean,
		val worldTemplate: JsonObject,
		val centralStructure: WorldStructure,
		val protectedDungeonIds: IntArraySet,
		val worldProperties: JsonObject,
		val spawningEnabled: Boolean
	)

	companion object {
		private val LOGGER = LogManager.getLogger()

		fun create(server: StarboundServer, template: WorldTemplate, storage: WorldStorage): ServerWorld {
			return ServerWorld(server, template, storage)
		}

		fun create(server: StarboundServer, geometry: WorldGeometry, storage: WorldStorage): ServerWorld {
			return create(server, WorldTemplate(geometry), storage)
		}

		fun load(server: StarboundServer, storage: WorldStorage): CompletableFuture<ServerWorld> {
			return storage.loadMetadata().thenApply {
				AssetPathStack("/") { _ ->
					val meta = it.map { Starbound.gson.fromJson(it.data.content, MetadataJson::class.java) }.orThrow { NoSuchElementException("No world metadata is present") }

					val world = create(server, WorldTemplate.fromJson(meta.worldTemplate), storage)
					world.playerSpawnPosition = meta.playerStart
					world.respawnInWorld = meta.respawnInWorld
					world.adjustPlayerSpawn = meta.adjustPlayerStart
					world.centralStructure = meta.centralStructure
					world.protectedDungeonIDs.addAll(meta.protectedDungeonIds)
					world
				}
			}
		}
	}
}
