package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.math.divideUp

fun positiveModulo(a: Int, b: Int): Int {
	val result = a % b
	return if (result < 0) result + b else result
}

fun positiveModulo(a: Double, b: Int): Double {
	val result = a % b
	return if (result < 0.0) result + b else result
}

fun positiveModulo(a: Double, b: Double): Double {
	val result = a % b
	return if (result < 0.0) result + b else result
}

fun positiveModulo(a: Float, b: Int): Float {
	val result = a % b
	return if (result < 0f) result + b else result
}

abstract class CoordinateMapper {
	abstract val chunks: Int

	abstract fun cell(value: Int): Int
	abstract fun cell(value: Double): Double
	abstract fun cell(value: Float): Float
	abstract fun chunk(value: Int): Int

	abstract fun chunkFromCell(value: Int): Int
	fun chunkFromCell(value: Float): Int = chunkFromCell(value.toInt())
	fun chunkFromCell(value: Double): Int = chunkFromCell(value.toInt())

	data class SplitResult(val rangeA: Vector2i, val rangeB: Vector2i?, val offset: Int)

	/**
	 * One of:
	 * * returns provided range as-is if it is contained in this range
	 * * splits in two if it wraps around one of edges
	 * * clamps if _this_ range is contained in provided range
	 */
	fun split(range: IntRange): SplitResult {
		return split(range.first, range.last)
	}

	abstract fun split(first: Int, last: Int): SplitResult

	// inside world bounds
	abstract fun inBoundsCell(value: Int): Boolean
	abstract fun inBoundsChunk(value: Int): Boolean

	// whenever provided index is legal (if world is wrapped around then provided index is legal outside world bounds)
	open fun isValidCellIndex(value: Int): Boolean = inBoundsCell(value)
	open fun isValidChunkIndex(value: Int): Boolean = inBoundsChunk(value)

	class Wrapper(private val cells: Int) : CoordinateMapper() {
		override val chunks = divideUp(cells, CHUNK_SIZE)

		override fun inBoundsCell(value: Int) = value in 0 until cells
		override fun inBoundsChunk(value: Int) = value in 0 until chunks

		override fun isValidCellIndex(value: Int) = true
		override fun isValidChunkIndex(value: Int) = true

		override fun chunkFromCell(value: Int): Int {
			return chunk(value shr CHUNK_SIZE_BITS)
		}

		override fun cell(value: Int): Int = positiveModulo(value, cells)
		override fun cell(value: Double): Double = positiveModulo(value, cells)
		override fun cell(value: Float): Float = positiveModulo(value, cells)
		override fun chunk(value: Int): Int = positiveModulo(value, chunks)

		override fun split(first: Int, last: Int): SplitResult {
			if (first >= last) {
				// point or empty range
				val wrap = cell(first)
				return SplitResult(Vector2i(wrap, wrap), null, 0)
			} else if (first <= 0 && last >= cells) {
				// covers entire world along this axis
				return SplitResult(Vector2i(0, cells - 1), null, 0)
			} else if (first >= 0 && last < cells) {
				// within range along this axis
				return SplitResult(Vector2i(first, last), null, 0)
			} else {
				val newFirst = cell(first)
				val newLast = cell(last)

				if (first < 0) {
					// wrapped around left edge

					return SplitResult(
						Vector2i(0, newLast),
						Vector2i(newFirst, cells - 1),
						newFirst - cells
					)
				} else {
					// wrapped around right edge

					return SplitResult(
						Vector2i(newFirst, cells - 1),
						Vector2i(0, newLast),
						-newLast - 1
					)
				}
			}
		}
	}

	class Clamper(private val cells: Int) : CoordinateMapper() {
		override val chunks = divideUp(cells, CHUNK_SIZE)

		override fun inBoundsCell(value: Int): Boolean {
			return value in 0 until cells
		}

		override fun chunkFromCell(value: Int): Int {
			return chunk(value shr CHUNK_SIZE_BITS)
		}

		override fun inBoundsChunk(value: Int): Boolean {
			return value in 0 until chunks
		}

		override fun split(first: Int, last: Int): SplitResult {
			// just clamp
			return SplitResult(Vector2i(cell(first), cell(last)), null, 0)
		}

		override fun cell(value: Int): Int {
			return value.coerceIn(0, cells - 1)
		}

		override fun cell(value: Double): Double {
			return value.coerceIn(0.0, cells - 1.0)
		}

		override fun cell(value: Float): Float {
			return value.coerceIn(0f, cells - 1f)
		}

		override fun chunk(value: Int): Int {
			return value.coerceIn(0, chunks - 1)
		}
	}
}
